# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import re
from libopensesame.oslogging import oslogger
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'SymbolSelector', category=u'extension')


PYTHON_SYMBOLS = r'^[ \t]*(?P<type>def|class)[ \t]+(?P<name>\w+)'
R_SYMBOLS = r'^[ \t]*(?P<name>[\w.]+)[ \t]*<-[ \t]*function'
MARKDOWN_HEADINGS = r'^#+[ \t]*(?P<name>.+)$'
MARKDOWN_HR = r'^---[ \t]*\n[ \t\n]*(?P<name>.{1,50})'


class SymbolSelector(BaseExtension):
    
    preferences_ui = 'extensions.SymbolSelector.preferences'

    def activate(self):

        self.main_window.set_busy(True)
        editor = self.extension_manager.provide(u'ide_current_editor')
        haystack = []
        if editor is not None and 'SymbolsMode' in editor.modes.keys():
            lsp_symbols = editor.modes.get('SymbolsMode').request_symbols()
            for name, kind, lineno, container in lsp_symbols:
                if container is not None:
                    name = '{}.{}'.format(container, name)
                haystack.append((
                    name,
                    lineno + 1,
                    self._jump_to_line
                ))
        else:
            mimetype = self.extension_manager.provide(u'ide_current_language')
            try:
                fnc = getattr(self, u'_get_{}_symbols'.format(mimetype))
            except AttributeError:
                oslogger.warning(u'don\'t know how to handle {}'.format(mimetype))
                return
            symbols = fnc(self.extension_manager.provide(u'ide_current_source'))
            for name, lineno in symbols:
                haystack.append((name, lineno, self._jump_to_line))
        self.main_window.set_busy(False)
        self.extension_manager.fire(
            u'quick_select',
            haystack=haystack,
            placeholder_text=_(u'Search symbols in current file …')
        )

    def _jump_to_line(self, lineno):

        self.extension_manager.fire('ide_jump_to_line', lineno=lineno)
        
    def _linenr(self, code, pos):
        
        return code[:pos].count('\n') + 1

    def _get_symbols(self, pattern, code):

        return [
            (m.group('name'), self._linenr(code, m.start()))
            for m in re.finditer(
                pattern,
                code,
                re.MULTILINE | re.ASCII if py3 else re.MULTILINE
            )
        ]

    def _get_nameless_symbols(self, pattern, code, tmpl):

        return [
            (tmpl.format(i + 1), self._linenr(code, m.start()))
            for i, m in enumerate(re.finditer(
                pattern,
                code,
                re.MULTILINE | re.ASCII if py3 else re.MULTILINE
            ))
        ]

    def _get_python_symbols(self, code):

        return self._get_symbols(PYTHON_SYMBOLS, code)

    def _get_R_symbols(self, code):

        return self._get_symbols(R_SYMBOLS, code)

    def _get_markdown_symbols(self, code):

        return (
            self._get_symbols(MARKDOWN_HEADINGS, code) +
            self._get_symbols(MARKDOWN_HR, code)
        )
        
    def _get_javascript_symbols(self, code):
        
        try:
            import esprima
        except ImportError:
            self.extension_manager.fire(
                'notify',
                message='Please install esprima for JavaScript support',
                category='warning'
            )
            return []
        from esprima.nodes import (
            ClassDeclaration,
            MethodDefinition,
            FunctionDeclaration
        )
        
        def parse_tree(ast):
            
            if isinstance(ast.body, list):
                symbols = []
                for e in ast.body:
                    symbols += parse_tree(e)
                return symbols
            if isinstance(
                ast.declaration,
                (ClassDeclaration, FunctionDeclaration)
            ):
                ast = ast.declaration
            if isinstance(ast, (ClassDeclaration, FunctionDeclaration)):
                return [(
                    ast.id.name,
                    self._linenr(code, ast.range[0])
                )] + parse_tree(ast.body)
            if isinstance(ast, MethodDefinition):
                return [(
                    ast.key.name,
                    self._linenr(code, ast.range[0])
                )] + parse_tree(ast.value.body)
            return []
        
        try:
            ast = esprima.parseScript(code, tolerant=True, range=True)
        except esprima.error_handler.Error:
            self.extension_manager.fire(
                'notify',
                message = 'Failed to parse JavaScript',
                category='warning'
            )
            return []
        return parse_tree(ast)

    def event_symbol_selector_activate(self):

        self.activate()



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import os
import uuid
import yaml
from qtpy import QtWidgets, QtCore
from pyqode.core import modes
from libopensesame.oslogging import oslogger
from libopensesame import misc
from libqtopensesame.misc.config import cfg
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'ImageAnnotations', category=u'extension')


class ImageAnnotations(BaseExtension):
    """Receives images that result from code execution (sent by JupyterConsole)
    and shows these as annotations next to the executed code. Relies on
    ImageAnnotationsMode and ImageAnnotationsPanel from pyqode.core.
    """
    
    def event_startup(self):
        """Creates a folder for the image files if it doesn't exist, and
        restores the image annotation info.
        """
        self._image_folder = os.path.join(
            self.main_window.home_folder,
            '.opensesame',
            'image_annotations'
        )
        if not os.path.exists(self._image_folder):
            oslogger.debug('creating {}'.format(self._image_folder))
            os.mkdir(self._image_folder)
        try:
            self._image_annotations = safe_yaml_load(cfg.image_annotations)
            assert(isinstance(self._image_annotations, dict))
        except BaseException:
            oslogger.warning('failed to parse image annotations from config')
            self._image_annotations = {}
            
    def event_register_editor(self, editor):
        """When an editor is registered, an ImageAnnotationsMode is added, but
        only if the editor already has an ImageAnnotationsPanel, which is
        managed by pyqode_manager.
        """
        try:
            editor.panels.get('ImageAnnotationsPanel')
        except KeyError:
            return
        mode = modes.ImageAnnotationsMode()
        mode.annotation_clicked.connect(self._on_annotation_clicked)
        editor.modes.append(mode)
        mode.set_annotations(self._image_annotations)
        
    def event_image_annotations_new(self, img, fmt, code):
        """Receives a new image annotation, where `img` is the image data,
        `fmt` is the format, and code is the code snippet that resulted in
        creating of the image.
        """
        editor = self.extension_manager.provide('ide_current_editor')
        if (
            editor is None or
            'ImageAnnotationsMode' not in list(editor.modes.keys())
        ):
            return
        if editor.file.path not in self._image_annotations:
            self._image_annotations[editor.file.path] = {}
        self._image_annotations[editor.file.path][code] \
            = self._figure_file(img, fmt)
        cfg.image_annotations = yaml.dump(self._image_annotations)
        editor.modes.get('ImageAnnotationsMode').set_annotations(
            self._image_annotations
        )
        
    def _figure_file(self, img, fmt):
        """Save an image to file and return the filename. Inspired by Spyder's
        `figurebrowser.py`.
        """
        if fmt == 'image/svg+xml':
            ext = '.svg'
        elif fmt == 'image/png':
            ext = '.png'
        elif fmt == 'image/jpeg':
            ext = '.jpg'
        else:
            raise ValueError('invalid image format: {}'.format(fmt))
        path = os.path.join(self._image_folder, uuid.uuid4().hex + ext)
        oslogger.debug('writing image to {}'.format(path))
        if fmt == 'image/svg+xml' and isinstance(img, str):
            img = img.encode('utf-8')
        with open(path, 'wb') as fd:
            fd.write(img)
        return path
    
    def _open_annotation(self, path):
        """Opens a single annotation image."""
        misc.open_url(path)
        
    def _open_folder(self):
        """Opens the folder that contains the image annotations."""
        misc.open_url(self._image_folder)
        
    def _clear_annotation(self, path=None):
        """Clear a single image annotation, or all image annotations in `path`
        is `None`.
        """
        editor = self.extension_manager.provide('ide_current_editor')
        if (
            editor is None or
            'ImageAnnotationsMode' not in list(editor.modes.keys())
        ):
            return
        if path is None:
            self._image_annotations[editor.file.path] = {}
        else:
            for code, img_path in list(
                self._image_annotations[editor.file.path].items()
            ):
                if img_path == path:
                    del self._image_annotations[editor.file.path][code]
        editor.modes.get('ImageAnnotationsMode').set_annotations(
            self._image_annotations
        )
        cfg.image_annotations = yaml.dump(self._image_annotations)

    def _on_annotation_clicked(self, msg, event):
        """Shows a context menu on the annotation."""
        if event.button() != QtCore.Qt.RightButton:
            return
        menu = QtWidgets.QMenu(self.main_window)
        menu.addAction(_('Open'), lambda: self._open_annotation(msg.path))
        menu.addSeparator()
        menu.addAction(_('Clear'), lambda: self._clear_annotation(msg.path))
        menu.addAction(_('Clear all'), self._clear_annotation)
        menu.addSeparator()
        menu.addAction(_('Open folder'), self._open_folder)
        menu.exec(event.globalPos())



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'WordCount', category=u'extension')


class WordCount(BaseExtension):

    def activate(self):

        source = self.extension_manager.provide('ide_current_selection')
        if not source:
            source = self.extension_manager.provide('ide_current_source')
        if source:
            message = _('{} lines, {} words, {} characters').format(
                source.count('\n') + 1,
                len(source.split()),
                len(source)
            )
        else:
            message = _('Nothing to count')
        self.extension_manager.fire(
            'notify',
            message=message,
            always_show=True
        )



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import importlib
from libopensesame.oslogging import oslogger
from qtpy.QtWidgets import QDockWidget
from qtpy.QtCore import Qt
from data_viewer_inspectors.inspect_str import inspect_str


class DataDockWidget(QDockWidget):
    
    def __init__(self, data_viewer, name, value, workspace):
        
        QDockWidget.__init__(self, data_viewer.main_window)
        self._data_viewer = data_viewer
        self._name = name
        self._workspace = workspace
        self.setObjectName('DataDockWidget_{}'.format(name))
        self.refresh(value)
        
    @property
    def workspace(self):
        
        return self._workspace
        
    def refresh(self, value):
        
        self._value = value
        cls = self._value.__class__.__name__
        fnc = self._inspect_fnc(cls)
        try:
            widget = fnc(value)
        except Exception as e:
            oslogger.debug('failed to inspect {}: {}'.format(self._name, e))
            widget = self._inspect_fallback(value)
        self.setWidget(widget)
        self.setWindowTitle('{}: {}'.format(self._name, cls))

    def _inspect_fnc(self, cls):

        try:
            m = importlib.import_module(
                'data_viewer_inspectors.inspect_{}'.format(cls)
            )
        except ModuleNotFoundError:
            oslogger.debug('using fallback inspector for type {}'.format(cls))
            return self._inspect_fallback
        else:
            oslogger.debug('using dedicated inspector for type {}'.format(cls))
        return getattr(m, 'inspect_{}'.format(cls))

    def _inspect_fallback(self, value):

        return inspect_str(repr(value))

    def closeEvent(self, e):

        self._data_viewer.remove_dock_widget(self._name)
        self.close()



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import os
import multiprocessing
from queue import Empty
import psutil
from qtpy.QtCore import Qt, QTimer
from qtpy.QtWidgets import QDockWidget
from datamatrix import DataMatrix
from qdatamatrix import QDataMatrix
from libqtopensesame.misc.config import cfg
from libopensesame.oslogging import oslogger
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'SubprocessManager', category=u'extension')

PARACHUTE_STOP = -1
PARACHUTE_HEARTBEAT = -2
PARACHUTE_HEARTBEAT_INTERVAL = 5


def parachute(queue, main_pid):
    
    """This function runs as a subprocess and receives the PIDs off all
    subprocesses. Whether the main process is alive is kept track of with a
    heartbeat. If no heartbeat comes in, then all subprocesses are killed to
    make sure that no runaway processes are left on shutdown.
    """

    oslogger.start('parachute')
    pids = [main_pid]
    oslogger.debug('main-process PID: {}'.format(main_pid))
    while True:
        try:
            pid = queue.get(True, 2 * PARACHUTE_HEARTBEAT_INTERVAL)
        except (Empty, ValueError, OSError):
            # If something is blocking the main process, the heartbeat stops.
            # This can happen for example if something is executed in an
            # in-process console. In that case we don't want to kill everything
            # off.
            if psutil.pid_exists(main_pid):
                oslogger.warning('main process is alive but unresponsive')
                continue
            oslogger.warning('main process appears to have died')
            break
        if pid == PARACHUTE_STOP:
            oslogger.debug('parchute stop')
            return
        if pid == PARACHUTE_HEARTBEAT:
            continue
        oslogger.debug('subprocess PID: {}'.format(pid))
        pids.append(pid)
    for pid in filter(psutil.pid_exists, pids):
        oslogger.warning('parachute killing PID: {}'.format(pid))
        p = psutil.Process(pid)
        p.kill()


class SubprocessDockWidget(QDockWidget):
    
    def __init__(self, subprocess_manager):
        
        QDockWidget.__init__(self, subprocess_manager.main_window)
        self._subprocess_manager = subprocess_manager
        self._qdm = QDataMatrix(self._subprocess_manager.dm(), read_only=True)
        self.setWidget(self._qdm)
        self.setWindowTitle(_('Subprocesses'))
        self.setObjectName('SubprocessManager')
        self.visibilityChanged.connect(self._refresh)
    
    def _refresh(self):

        if not self.isVisible():
            return
        self._qdm.dm = self._subprocess_manager.dm()
        self._qdm.refresh()
        QTimer.singleShot(2000, self._refresh)
    

class SubprocessManager(BaseExtension):
    
    def activate(self):

        if not hasattr(self, '_dock_widget'):
            self._dock_widget = SubprocessDockWidget(self)
            self.main_window.addDockWidget(
                Qt.RightDockWidgetArea,
                self._dock_widget
            )
        self._dock_widget.setVisible(True)

    def dm(self):
        
        died = []
        pids = [os.getpid()]
        states = ['running']
        descs = ['MainProcess']
        for pid, description in self._processes.items():
            if not psutil.pid_exists(pid):
                self._ended.append((pid, description))
                died.append(pid)
                continue
            pids.append(pid)
            states.append('running')
            descs.append(description)
        for pid in died:
            del self._processes[pid]
        if cfg.subprocess_manager_show_ended:
            for pid, description in self._ended:
                pids.append(pid)
                states.append('ended')
                descs.append(description)
        dm = DataMatrix(length=len(pids))
        dm.pid = pids
        dm.state = states
        dm.description = descs
        return dm

    def event_register_subprocess(self, pid, description):
        
        if not hasattr(self, '_parachute'):
            self._processes = {}
            self._parachute_start()
            self._ended = []
        oslogger.debug('{}: {}'.format(pid, description))
        self._processes[pid] = description
        self._queue.put(pid)

    def event_close(self):
        
        for pid in self._active_processes:
            oslogger.debug('killing process {}'.format(pid))
            p = psutil.Process(pid)
            p.kill()
            
    def event_run_experiment(self, fullscreen):
        
        self._parachute_stop()
        
    def event_end_experiment(self, ret_val):
        
        self._parachute_start()
        
    def provide_subprocess_pids(self):
        
        return list(self._processes.keys())
            
    def _parachute_start(self):
        
        self._queue = multiprocessing.Queue()
        self._parachute = multiprocessing.Process(
            target=parachute,
            args=(self._queue, os.getpid())
        )
        self._parachute.start()
        for pid in self._active_processes:
            self._queue.put(pid)
        self._processes[self._parachute.pid] = 'parachute'
        QTimer.singleShot(
            1000 * PARACHUTE_HEARTBEAT_INTERVAL,
            self._pararchute_heartbeat
        )
        
    @property
    def _active_processes(self):
        
        return filter(psutil.pid_exists, self._processes)
        
    def _parachute_stop(self):
        
        self._queue.put(PARACHUTE_STOP)

    def _pararchute_heartbeat(self):
        
        self._queue.put(PARACHUTE_HEARTBEAT)
        QTimer.singleShot(
            1000 * PARACHUTE_HEARTBEAT_INTERVAL,
            self._pararchute_heartbeat
        )



# coding=utf-8
import bdb
import sys
import inspect
import zmq
import pickle
import traceback
from IPython.core.debugger import Pdb as IPdb
from IPython.core.getipython import get_ipython


class RapunzelPdb(IPdb, object):

    def __init__(self, port=5555, breakpoints=None):

        super(RapunzelPdb, self).__init__()
        self.clear_all_breaks()
        self._set_breakpoints(breakpoints)
        self._zmq_context = zmq.Context()
        self._zmq_socket = self._zmq_context.socket(zmq.REQ)
        self._zmq_socket.connect('tcp://localhost:{}'.format(port))
        self._zmq_socket.send_pyobj({'type': 'pdb_start'})
        self.post_mortem = False
       
    def clear_all_breaks(self):
        
        # Adapted from SpyderPDB.set_spyder_break_points. For some reason you
        # need to work hard to avoid breakpoints from becoming persistent
        super(RapunzelPdb, self).clear_all_breaks()
        for bp in bdb.Breakpoint.bpbynumber:
            if bp:
                bp.deleteMe()
        bdb.Breakpoint.next = 1
        bdb.Breakpoint.bplist = {}
        bdb.Breakpoint.bpbynumber = [None]
        
    def _set_breakpoints(self, breakpoints):
        
        if not breakpoints:
            return
        for break_path, lines in breakpoints.items():
            for line in lines:
                self.set_break(break_path, line + 1)
                
    def _transmit_workspace(self, frame):
        
        # # See transparent_jupyter_widget
        workspace_variables = frame.f_locals
        workspace_variables.update(frame.f_globals)
        workspace_variables = ({
            key: (
                repr(val),
                val.__class__.__name__,
                (
                    val.shape if hasattr(val, 'shape')
                    else len(val) if hasattr(val, '__len__')
                    else None
                )
            )
            for key, val in workspace_variables.items()
            if not key.startswith(u'_') and
            key not in ('In', 'Out') and
            not callable(val) and
            not inspect.isclass(val) and
            not inspect.ismodule(val)
        })
        try:
            self._zmq_socket.send_pyobj({
                'type': 'pdb_frame',
                'path': frame.f_code.co_filename,
                'line': frame.f_lineno,
                'workspace': workspace_variables
            })
        except zmq.error.ZMQError:
            pass
        try:
            self._zmq_socket.recv()
        except zmq.error.ZMQError:
            pass
        
    def close(self):
        
        self._zmq_socket.send_pyobj({'type': 'pdb_stop'})
        self._zmq_socket.close()

    def interaction(self, frame, traceback):
        
        self._transmit_workspace(frame)
        super(RapunzelPdb, self).interaction(frame, traceback)

    def postcmd(self, stop, line):
        
        self._transmit_workspace(self.curframe)
        return super(RapunzelPdb, self).postcmd(stop, line)


def runfile(path):

    with open(path) as f:
        code = f.read()
    bytecode = compile(code, path, 'exec')
    exec(bytecode, {})


def rpdb(path, port=5555, breakpoints=None):

    print('''
* Enter `c` (continue) to start execution
* Enter `q` (quit) to quit
* Enter `?` for help
''')
    # Ideally we'd set the first breakpoint to line 1 and then execute until
    # there. However, this is tricky to accomplish.
    # if not breakpoints or all(not lines for lines in breakpoints.values()):
    #     print('Settig first breakpoint at line 1')
    #     breakpoints = {path: [0]}
    pdb = RapunzelPdb(port, breakpoints)
    try:
        pdb.run('runfile({})'.format(repr(path)))
    except Exception as e:
        tb = sys.exc_info()[2]
        traceback.print_exception(type(e), e, tb)
        pdb.reset()
        pdb.post_mortem = True
        while tb.tb_next:  # Get the highest frame (where the error occurred)
            tb = tb.tb_next
        pdb.interaction(tb.tb_frame, tb)
    finally:
        pdb.close()  # Close connection also on crash



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import uuid
import os
import ast
import time
import json
import inspect
import pickle
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from qtpy.QtWidgets import QApplication
from libopensesame.oslogging import oslogger
from qtconsole.rich_jupyter_widget import RichJupyterWidget
try:
    from base64 import decodebytes
except ImportError:  # Python 2
    from base64 import decodestring as decodebytes
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'JupyterConsole', category=u'extension')


GLOBAL_EXPR = u'''{
key: (
    repr(val),
    val.__class__.__name__,
    (
        val.shape if hasattr(val, 'shape')
        else len(val) if hasattr(val, '__len__')
        else None
    )
)
for key, val in globals().items()
if not key.startswith(u'_') and
key not in ('In', 'Out') and
not callable(val) and
not inspect.isclass(val) and
not inspect.ismodule(val)
}'''

LIST_GLOBAL_EXPR = u'''[
key
for key, val in globals().items()
if not key.startswith(u'_') and
key not in ('In', 'Out') and
not callable(val) and
not inspect.isclass(val) and
not inspect.ismodule(val)
]
'''


class FailedToGetWorkspaceVariable(Exception):
    
    def __repr__(self):
        
        return _('Failed to get workspace variable')


class TransparentJupyterWidget(RichJupyterWidget, BaseSubcomponent):

    def __init__(self, jupyter_console):

        self._name = jupyter_console.name
        self._jupyter_console = jupyter_console
        super(TransparentJupyterWidget, self).__init__(jupyter_console)
        self.setup(jupyter_console)
        self.executed.connect(self._on_executed)
        self.executing.connect(self._on_executing)
        self._last_executed = None
        # It is possible for _on_executing() to be called again, before
        # _on_executed() is called. We therefore keep track of how many times
        # _on_executing() has been called, and only release the busy status
        # when this counter is set back to 0.
        self._executing_counter = 0
        
    def execute(self, *args, **kwargs):
        
        self._last_executed = args[0] if args else kwargs.get('source', None)
        super(TransparentJupyterWidget, self).execute(*args, **kwargs)

    def _on_executing(self):

        self._executing_counter += 1
        self._jupyter_console.set_busy(True)

    def _on_executed(self):

        if self._executing_counter:
            self._executing_counter -= 1
        if self._executing_counter:
            return
        self._jupyter_console.set_busy(False)
        self.extension_manager.fire(
            u'workspace_update',
            name=self._name,
            workspace_func=self.get_workspace_globals
        )
        
    def running(self):
        
        return self._executing_counter > 0
    
    def reset(self, clear=False):
        
        super(TransparentJupyterWidget, self).reset(clear=clear)
        self._executing_counter = 0
        self._jupyter_console.set_busy(False)
        
    def _handle_display_data(self, msg):
        """Reimplemented to handle communications between the figure explorer
        and the kernel. Inspired by Spyder's `figurebrowser.py`.
        """
        img = None
        data = msg['content']['data']
        if 'image/svg+xml' in data:
            fmt = 'image/svg+xml'
            img = data['image/svg+xml']
        elif 'image/png' in data:
            # PNG data is base64 encoded as it passes over the network
            # in a JSON structure so we decode it.
            fmt = 'image/png'
            img = decodebytes(data['image/png'].encode('ascii'))
        elif 'image/jpeg' in data and self._jpg_supported:
            fmt = 'image/jpeg'
            img = decodebytes(data['image/jpeg'].encode('ascii'))
        if img is not None:
            self.extension_manager.fire(
                'image_annotations_new',
                img=img,
                fmt=fmt,
                code=self._last_executed
            )
        return super(TransparentJupyterWidget, self)._handle_display_data(msg)


class OutprocessJupyterWidget(TransparentJupyterWidget):

    """Makes the Python workspace of a Jupyter console with a kernel running in
    a different process accessible.
    """

    def __init__(self, jupyter_console):

        self._user_expressions = {}
        super(OutprocessJupyterWidget, self).__init__(jupyter_console)

    def _handle_execute_reply(self, msg):

        self._user_expressions = msg.get(
            u'content',
            {}
        ).get(u'user_expressions', {})
        return super(OutprocessJupyterWidget, self)._handle_execute_reply(msg)
        
    def _silent_execute(self, expr):
        
        key = str(uuid.uuid4())
        self._kernel_client.execute(
            u'import os, inspect',
            silent=True,
            user_expressions={
                key: expr
            }
        )
        for _ in range(100):
            if key in self._user_expressions:
                break
            time.sleep(0.01)
            QApplication.processEvents()
        else:
            return {u'no reply': None}
        reply = self._user_expressions[key].get(u'data', {}).get(
            u'text/plain',
            u'{"invalid reply": None}'
        )
        try:
            return ast.literal_eval(reply)
        except (ValueError, SyntaxError):
            return {u'cannot eval reply': None}

    def get_workspace_globals(self):

        return self._silent_execute(GLOBAL_EXPR)
        
    @property
    def pid(self):
        
        for i in range(10):
            pid = self._silent_execute('os.getpid()')
            if isinstance(pid, int):
                return pid
            time.sleep(0.1)
        oslogger.warning('failed to get pid')
        return -1

    def list_workspace_globals(self):

        return self._silent_execute(LIST_GLOBAL_EXPR)

    def set_workspace_globals(self, global_dict):

        code = ['from pickle import loads']
        for var, val in global_dict.items():
            if (
                var.startswith('_') or
                callable(val) or
                inspect.isclass(val) or
                inspect.ismodule(val)
            ):
                continue
            try:
                blob = pickle.dumps(val)
            except (pickle.PicklingError, TypeError):
                pass
            else:
                code.append('{} = loads({})'.format(var, repr(blob)))
        self._kernel_client.execute(';'.join(code), silent=True)

    def get_workspace_variable(self, name):

        key = str(uuid.uuid4())
        self._kernel_client.execute(
            u'import pickle',
            silent=True,
            user_expressions={
                key: u'pickle.dumps({})'.format(name)
            }
        )
        for _ in range(200):
            if key in self._user_expressions:
                break
            time.sleep(0.05)
            QApplication.processEvents()
        else:
            return FailedToGetWorkspaceVariable()
        reply = self._user_expressions[key].get(
            u'data',
            {}
        ).get(u'text/plain', None)
        if reply is None:
            return FailedToGetWorkspaceVariable()
        try:
            return pickle.loads(eval(reply))
        except (
            TypeError,
            pickle.UnpicklingError,
            RecursionError,
            ModuleNotFoundError
        ):
            return FailedToGetWorkspaceVariable()


class InprocessJupyterWidget(TransparentJupyterWidget):

    """Makes the Python workspace of a Jupyter console with a kernel running in
    the same process accessible.
    """

    def __init__(self, parent):

        self._jupyter_console = parent
        super(InprocessJupyterWidget, self).__init__(parent)

    def get_workspace_globals(self):

        return {
            key: (
                json.dumps(val, default=lambda x: u'<no preview>'),
                val.__class__.__name__,
                len(val) if hasattr(val, '__len__') else '<na>'
            )
            for key, val
            in self._kernel_manager.kernel.shell.user_global_ns.copy().items()
            if not key.startswith(u'_') and
            key not in ('In', 'Out') and
            not callable(val) and
            not inspect.isclass(val) and
            not inspect.ismodule(val)
        }
    
    @property
    def pid(self):

        return os.getpid()
        
    def list_workspace_globals(self):
        
        return [
            key
            for key, val
            in self._kernel_manager.kernel.shell.user_global_ns.copy().items()
            if not key.startswith(u'_') and
            key not in ('In', 'Out') and
            not callable(val) and
            not inspect.isclass(val) and
            not inspect.ismodule(val)
        ]

    def set_workspace_globals(self, global_dict):

        self._kernel_manager.kernel.shell.push(global_dict)

    def get_workspace_variable(self, name):

        return self._kernel_manager.kernel.shell.user_global_ns.get(name, None)
